# pylint: disable=too-many-locals
from typing import Dict, List

import pytest
from pytest_mock import MockFixture
from pytest_test_utils import TmpDir
from scmrepo.git import Git

from gto.registry import GitRegistry

from .utils import check_obj

EXPECTED_REGISTRY_TAG_TAG_STATE = {
    "artifacts": {
        "rf": {
            "artifact": "rf",
            "versions": [
                {
                    "artifact": "rf",
                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                    "version": "v1.2.3",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "rf",
                            "created_at": "2022-08-04T16:56:57",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Add artifacts",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "version": "v1.2.3",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "model",
                                        "path": "models/random-forest.pkl",
                                        "virtual": False,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [
                        {
                            "priority": 3,
                            "addition": True,
                            "artifact": "rf",
                            "created_at": "2022-08-04T16:56:57",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Registering artifact rf version v1.2.3",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "tag": "rf@v1.2.3",
                            "version": "v1.2.3",
                        }
                    ],
                    "deregistrations": [],
                    "stages": {
                        "production": {
                            "artifact": "rf",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "version": "v1.2.3",
                            "stage": "production",
                            "assignments": [
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "rf",
                                    "created_at": "2022-08-04T16:56:59",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage production to artifact rf version v1.2.3",
                                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                                    "tag": "rf#production#1",
                                    "version": "v1.2.3",
                                    "stage": "production",
                                },
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "rf",
                                    "created_at": "2022-08-04T16:57:02",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage production to artifact rf version v1.2.3",
                                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                                    "tag": "rf#production#4",
                                    "version": "v1.2.3",
                                    "stage": "production",
                                },
                            ],
                            "unassignments": [],
                        }
                    },
                },
                {
                    "artifact": "rf",
                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                    "version": "v1.2.4",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "rf",
                            "created_at": "2022-08-04T16:56:59",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Update model",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "v1.2.4",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "model",
                                        "path": "models/random-forest.pkl",
                                        "virtual": False,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [
                        {
                            "priority": 3,
                            "addition": True,
                            "artifact": "rf",
                            "created_at": "2022-08-04T16:56:59",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Registering artifact rf version v1.2.4",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "tag": "rf@v1.2.4",
                            "version": "v1.2.4",
                        }
                    ],
                    "deregistrations": [],
                    "stages": {
                        "staging": {
                            "artifact": "rf",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "v1.2.4",
                            "stage": "staging",
                            "assignments": [
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "rf",
                                    "created_at": "2022-08-04T16:57:00",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage staging to artifact rf version v1.2.4",
                                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                                    "tag": "rf#staging#2",
                                    "version": "v1.2.4",
                                    "stage": "staging",
                                }
                            ],
                            "unassignments": [],
                        },
                        "production": {
                            "artifact": "rf",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "v1.2.4",
                            "stage": "production",
                            "assignments": [
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "rf",
                                    "created_at": "2022-08-04T16:57:01",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage production to artifact rf version v1.2.4",
                                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                                    "tag": "rf#production#3",
                                    "version": "v1.2.4",
                                    "stage": "production",
                                }
                            ],
                            "unassignments": [],
                        },
                    },
                },
            ],
            "creations": [],
            "deprecations": [],
        },
        "nn": {
            "artifact": "nn",
            "versions": [
                {
                    "artifact": "nn",
                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                    "version": "v0.0.1",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "nn",
                            "created_at": "2022-08-04T16:56:57",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Add artifacts",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "version": "v0.0.1",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "model",
                                        "path": "models/neural-network.pkl",
                                        "virtual": False,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [
                        {
                            "priority": 3,
                            "addition": True,
                            "artifact": "nn",
                            "created_at": "2022-08-04T16:56:58",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Registering artifact nn version v0.0.1",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "tag": "nn@v0.0.1",
                            "version": "v0.0.1",
                        }
                    ],
                    "deregistrations": [],
                    "stages": {
                        "staging": {
                            "artifact": "nn",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "version": "v0.0.1",
                            "stage": "staging",
                            "assignments": [
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "nn",
                                    "created_at": "2022-08-04T16:56:59",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage staging to artifact nn version v0.0.1",
                                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                                    "tag": "nn#staging#1",
                                    "version": "v0.0.1",
                                    "stage": "staging",
                                }
                            ],
                            "unassignments": [],
                        }
                    },
                },
                {
                    "artifact": "nn",
                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                    "version": "04d79900801d9aa7ec726169706280a32a25d198",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "nn",
                            "created_at": "2022-08-04T16:56:59",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Update model",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "04d79900801d9aa7ec726169706280a32a25d198",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "model",
                                        "path": "models/neural-network.pkl",
                                        "virtual": False,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [],
                    "deregistrations": [],
                    "stages": {},
                },
            ],
            "creations": [],
            "deprecations": [],
        },
        "features": {
            "artifact": "features",
            "versions": [
                {
                    "artifact": "features",
                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                    "version": "04d79900801d9aa7ec726169706280a32a25d198",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "features",
                            "created_at": "2022-08-04T16:56:59",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Update model",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "04d79900801d9aa7ec726169706280a32a25d198",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "dataset",
                                        "path": "datasets/features.csv",
                                        "virtual": True,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [],
                    "deregistrations": [],
                    "stages": {},
                }
            ],
            "creations": [],
            "deprecations": [],
        },
    }
}


def iter_over(sequence):
    if isinstance(sequence, dict):
        return list(sequence.values())
    if isinstance(sequence, list):
        return sequence
    raise NotImplementedError


@pytest.mark.usefixtures("showcase")
def test_registry_state_tag_tag(tmp_dir: TmpDir):
    with GitRegistry.from_url(tmp_dir) as reg:
        appeared_state = reg.get_state().dict()

    # TODO: update state
    exclude: Dict[str, List[str]] = {
        "commits": [],
        "versions": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "registrations",
            "deregistrations",
            "enrichments",
            "stages",
            "message",
            "version",
        ],
        "enrichments": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "message",
            "version",
            "committer",
            "committer_email",
        ],
        "registrations": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "message",
            "version",
        ],
        "deregistrations": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "message",
            "version",
        ],
        "stages": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "message",
            "version",
            "assignments",
            "unassignments",
        ],
    }

    expected_state = EXPECTED_REGISTRY_TAG_TAG_STATE
    for artifact in expected_state["artifacts"]:
        for appeared, expected in zip(
            iter_over(appeared_state["artifacts"][artifact]["versions"]),
            iter_over(expected_state["artifacts"][artifact]["versions"]),
        ):
            check_obj(appeared, expected, exclude["versions"])

            for key in ["enrichments", "registrations", "deregistrations"]:
                for a, e in zip(
                    iter_over(appeared[key]),
                    iter_over(expected[key]),
                ):
                    check_obj(a, e, exclude[key])

            for key in appeared["stages"]:
                check_obj(
                    appeared["stages"][key], expected["stages"][key], exclude["stages"]
                )


def test_from_url_sets_cloned_property(tmp_dir: TmpDir, scm: Git, mocker: MockFixture):
    with GitRegistry.from_url(tmp_dir) as reg:
        assert reg.cloned is False

    with GitRegistry.from_url(scm) as reg:
        assert reg.cloned is False

    cloned_git_repo_mock = mocker.patch("gto.git_utils.cloned_git_repo")
    cloned_git_repo_mock.return_value.__enter__.return_value = scm

    with GitRegistry.from_url("https://github.com/iterative/gto") as reg:
        assert reg.cloned is True


# Some method parameters (model names, versions, revs, etc) depend and set by
# the `showcase` fixture setup in the conftest.py.
@pytest.mark.parametrize(
    "method,args,kwargs",
    [
        ("register", ["new_model", "HEAD"], {}),
        ("deregister", ["nn"], {"version": "v0.0.1"}),
        ("assign", ["nn", "new_stage"], {"version": "v0.0.1"}),
        ("unassign", ["nn", "staging"], {"version": "v0.0.1"}),
        ("deprecate", ["nn"], {}),
    ],
)
@pytest.mark.usefixtures("showcase")
def test_tag_is_pushed_if_cloned_is_set(
    tmp_dir: TmpDir,
    mocker: MockFixture,
    method,
    args,
    kwargs,
):
    with GitRegistry.from_url(tmp_dir) as reg:
        # imitate that we are doing actions on the remote repo
        assert reg.cloned is False
        reg.cloned = True

        # check that it attempts to push tag to a remote repo, even if
        # push=False is set in a call. `cloned` overrides it in this case
        git_push_tag_mock = mocker.patch("gto.registry.git_push_tag")
        kwargs["push"] = False
        getattr(reg, method)(*args, **kwargs)
        git_push_tag_mock.assert_called_once()



import logging
import os
from contextlib import contextmanager
from typing import List, Optional, TypeVar, cast

from funcy import distinct
from scmrepo.git import Git

from gto.base import (
    Assignment,
    BaseEvent,
    BaseRegistryState,
    Deprecation,
    Deregistration,
    Registration,
    Unassignment,
)
from gto.config import CONFIG_FILE_NAME, RegistryConfig, read_registry_config
from gto.constants import NAME, assert_fullname_is_valid
from gto.exceptions import (
    InvalidTagName,
    NotImplementedInGTO,
    VersionAlreadyRegistered,
    VersionExistsForCommit,
    WrongArgs,
)
from gto.git_utils import RemoteRepoMixin, git_push_tag
from gto.index import EnrichmentManager
from gto.tag import (
    TagArtifactManager,
    TagStageManager,
    TagVersionManager,
    delete_tag,
    parse_name,
)
from gto.ui import echo
from gto.versions import SemVer

from ._pydantic import BaseModel

TBaseEvent = TypeVar("TBaseEvent", bound=BaseEvent)


class GitRegistry(BaseModel, RemoteRepoMixin):
    scm: Git
    cloned: bool
    artifact_manager: TagArtifactManager
    version_manager: TagVersionManager
    stage_manager: TagStageManager
    enrichment_manager: EnrichmentManager
    config: RegistryConfig

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    @contextmanager
    def from_scm(
        cls,
        scm: Git,
        cloned: bool = False,
        config: Optional[RegistryConfig] = None,
    ):
        if config is None:
            config = read_registry_config(os.path.join(scm.root_dir, CONFIG_FILE_NAME))

        yield cls(
            scm=scm,
            cloned=cloned,
            config=config,
            artifact_manager=TagArtifactManager(scm=scm, config=config),
            version_manager=TagVersionManager(scm=scm, config=config),
            stage_manager=TagStageManager(scm=scm, config=config),
            enrichment_manager=EnrichmentManager(scm=scm, config=config),
        )

    def is_gto_repo(self):
        if self.config.config_file_exists():
            return True
        if self.config.check_index_exist(self.scm.root_dir):
            return True
        if self.get_state() != BaseRegistryState():
            return True
        return False

    def get_state(
        self,
        all_branches=False,
        all_commits=False,
    ) -> BaseRegistryState:
        state = BaseRegistryState()
        state = self.artifact_manager.update_state(state)
        state = self.version_manager.update_state(state)
        state = self.stage_manager.update_state(state)
        state = self.enrichment_manager.update_state(
            state,
            all_branches=all_branches,
            all_commits=all_commits,
        )
        return state

    def get_artifacts(
        self,
        all_branches=False,
        all_commits=False,
    ):
        return self.get_state(
            all_branches=all_branches,
            all_commits=all_commits,
        ).get_artifacts()

    def find_artifact(
        self,
        name: Optional[str] = None,
        create_new=False,
        all_branches=False,
        all_commits=False,
    ):
        return self.get_state(
            all_branches=all_branches,
            all_commits=all_commits,
        ).find_artifact(
            name, create_new=create_new  # type: ignore
        )

    def register(  # pylint: disable=too-many-locals
        self,
        name,
        rev,
        version=None,
        message=None,
        simple=None,
        force=False,
        bump_major=False,
        bump_minor=False,
        bump_patch=False,
        push=False,
        stdout=False,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> Registration:
        """Register artifact version"""
        assert_fullname_is_valid(name)
        version_args = sum(
            bool(i) for i in (version, bump_major, bump_minor, bump_patch)
        )
        if version_args > 1:
            raise WrongArgs("Need to specify either version or single bump argument")
        rev = self.scm.resolve_rev(rev)
        found_artifact = self.find_artifact(name, create_new=True)
        # check that this commit don't have a version already
        found_version = found_artifact.find_version(commit_hexsha=rev)
        if found_version is not None:
            if not force and found_version.is_registered and found_version.is_active:
                raise VersionExistsForCommit(name, found_version.version)
            if not version:
                version = found_version.version
            elif found_version.version != version:
                raise WrongArgs(
                    f"For this REF you can only register {found_version.version}"
                )
        # if version name is provided, use it
        if version:
            SemVer(version)
            if found_artifact.find_version(name=version) != found_version:
                raise VersionAlreadyRegistered(version)
        else:
            # if version name wasn't provided but there were some, bump the last one
            last_version = found_artifact.get_latest_version(registered_only=True)
            if last_version:
                version = (
                    SemVer(last_version.version)
                    .bump(
                        bump_major=bump_major,
                        bump_minor=bump_minor,
                        bump_patch=bump_patch if (bump_major or bump_minor) else True,
                    )
                    .version
                )
            elif version_args:
                raise WrongArgs(
                    "Can't apply bump, because this is the first version being registered"
                )
            else:
                version = SemVer.get_minimal().version
        tag = self.version_manager.register(
            name,
            version,
            rev,
            message=message or f"Registering artifact {name} version {version}",
            simple=simple,
            author=author,
            author_email=author_email,
        )
        if stdout:
            echo(f"Created git tag '{tag}' that registers version")
        self._push_tag_or_echo_reminder(tag_name=tag, push=push, stdout=stdout)
        return self._return_event(tag)

    def deregister(  # pylint: disable=too-many-locals
        self,
        name,
        rev=None,
        version=None,
        message=None,
        simple=None,
        force=False,
        delete=False,
        push: bool = False,
        stdout=False,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> Deregistration:
        """Deregister artifact version"""
        self._check_args(name, version, rev)
        if rev is not None:
            rev = self.scm.resolve_rev(rev)
        found_artifact = self.find_artifact(name, create_new=True)
        if not force:
            found_version = found_artifact.find_version(
                name=version, commit_hexsha=rev, raise_if_not_found=True
            )
            if not found_version.is_registered:
                raise WrongArgs(
                    f"The version at ref '{found_version.commit_hexsha}' is not registered"
                )
            if not found_version.is_active:
                raise WrongArgs(
                    f"The version at ref '{found_version.commit_hexsha}' was deregistered already"
                )

        found_version = found_artifact.find_version(name=version, commit_hexsha=rev)
        version = version or getattr(found_version, "version", None)
        commit_hexsha = rev or getattr(found_version, "commit_hexsha", None)
        if not (version and commit_hexsha):
            raise WrongArgs("Can't deregister a version that have no git tags")

        if delete:
            tags = distinct(
                [
                    e.tag
                    for e in found_version.get_events(ascending=True)
                    if hasattr(e, "tag")
                ]
            )
            return self._delete_tags(tags, stdout=stdout, push=push)

        tag = self.version_manager.deregister(
            name,
            version,
            found_version.commit_hexsha,
            message=message or f"Deregistering artifact {name} version {version}",
            simple=simple,
            author=author,
            author_email=author_email,
        )
        if stdout:
            echo(f"Created git tag '{tag}' that deregisters version")
        self._push_tag_or_echo_reminder(
            tag_name=tag, push=push, stdout=stdout, delete=delete
        )
        return self._return_event(tag)

    def assign(  # pylint: disable=too-many-locals
        self,
        name,
        stage,
        version=None,
        rev=None,
        name_version=None,
        message=None,
        simple=False,
        force=False,
        push: bool = False,
        skip_registration=False,
        stdout=False,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> Assignment:
        """Assign stage to specific artifact version"""
        self._check_args(name, version, rev, stage)
        if name_version:
            self._check_version(name_version)
        if name_version and skip_registration:
            raise WrongArgs(
                "You either need to supply version name or skip registration"
            )
        if rev:
            rev = self.scm.resolve_rev(rev)
        found_artifact = self.find_artifact(name, create_new=True)
        if version:
            found_version = found_artifact.find_version(
                name=version, raise_if_not_found=False
            )
            if not found_version:
                raise WrongArgs(f"Version '{version}' is not registered")
            rev = self.find_commit(name, version)
        else:
            found_version = found_artifact.find_version(commit_hexsha=rev)
            if found_version:
                if name_version:
                    raise WrongArgs(
                        f"Can't register '{SemVer(name_version).version}', since '{found_version.version}' is registered already at this ref"
                    )
            else:
                if not skip_registration:
                    self.register(
                        name,
                        version=name_version,
                        rev=rev,
                        simple=True,
                        stdout=stdout,
                        push=push,
                    )
                found_version = self.find_artifact(name, create_new=True).find_version(
                    commit_hexsha=rev, create_new=True
                )
        if (
            not force
            and found_version
            and found_version.stages
            and any(vstage.stage == stage for vstage in found_version.get_vstages())
        ):
            raise WrongArgs(
                f"Version '{found_version.version}' is already in stage '{stage}'. Use the '--force' flag to create a new Git tag nevertheless."
            )
        # TODO: getting tag name as a result and using it
        # is leaking implementation details in base module
        # it's roughly ok to have until we add other implementations
        # beside tag-based assignments
        tag = self.stage_manager.assign(  # type: ignore
            name,
            stage,
            rev=rev,
            message=message
            or f"Assigning stage {stage} to artifact {name} "
            + (
                f"version {found_version.version}" if found_version else f"commit {rev}"
            ),
            simple=simple,
            author=author,
            author_email=author_email,
        )
        if stdout:
            echo(
                f"Created git tag '{tag}' that assigns stage to version '{found_version.version}'"
            )
        self._push_tag_or_echo_reminder(tag_name=tag, push=push, stdout=stdout)
        return self._return_event(tag)

    def unassign(  # pylint: disable=too-many-locals
        self,
        name,
        stage,
        version=None,
        rev=None,
        message=None,
        simple=False,
        force=False,
        delete=False,
        push: bool = False,
        stdout=False,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> Unassignment:
        """Unassign stage to specific artifact version"""
        self._check_args(name, version, rev, stage)
        if rev:
            rev = self.scm.resolve_rev(rev)
        found_artifact = self.find_artifact(name)
        found_version = found_artifact.find_version(
            name=version, commit_hexsha=rev, raise_if_not_found=True
        )
        if not force and all(s != stage for s in getattr(found_version, "stages", [])):
            raise WrongArgs(
                f"Stage '{stage}' is not assigned to a version '{found_version.version}'"
            )
        if delete:
            tags = distinct(
                [
                    e.tag
                    for e in found_version.get_events(ascending=True)
                    if hasattr(e, "tag") and hasattr(e, "stage") and e.stage == stage
                ]
            )
            return self._delete_tags(tags, push=push, stdout=stdout)

        # TODO: getting tag name as a result and using it
        # is leaking implementation details in base module
        # it's roughly ok to have until we add other implementations
        # beside tag-based assignments
        tag = self.stage_manager.unassign(  # type: ignore
            name,
            stage,
            rev=found_version.commit_hexsha,
            message=message
            or f"Unassigning stage '{stage}' to artifact '{name}' version '{found_version.version}'",
            simple=simple,
            author=author,
            author_email=author_email,
        )
        if stdout:
            echo(
                f"Created git tag '{tag}' that unassigns stage from version '{found_version.version}'"
            )
        self._push_tag_or_echo_reminder(
            tag_name=tag, push=push, stdout=stdout, delete=delete
        )
        return self._return_event(tag)

    def deprecate(
        self,
        name,
        rev=None,
        message=None,
        simple=False,
        force=False,
        delete=False,
        push: bool = False,
        stdout=False,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> Optional[Deprecation]:
        if force:
            if simple:
                raise WrongArgs("Can't use 'force' with 'simple=True'")
            simple = False
        else:
            if simple is None:
                simple = True
            found_artifact = self.find_artifact(name)
            if not found_artifact.is_active:
                raise WrongArgs("Artifact was deprecated already")
        if delete:
            tags = distinct(
                [
                    e.tag
                    for e in self.find_artifact(name=name).get_events(ascending=True)
                    if hasattr(e, "tag")
                ]
            )
            return self._delete_tags(tags, stdout=stdout, push=push)
        if rev is None:
            if name in self.get_artifacts():
                rev = self.find_artifact(name=name).get_events()[0].commit_hexsha
            else:
                rev = "HEAD"
        tag = self.artifact_manager.deprecate(  # type: ignore
            name,
            rev=rev,
            message=message or f"Deprecating artifact '{name}'",
            simple=simple,
            author=author,
            author_email=author_email,
        )
        if stdout:
            echo(f"Created git tag '{tag}' that deprecates artifact")
        self._push_tag_or_echo_reminder(
            tag_name=tag, push=push, stdout=stdout, delete=delete
        )
        return self._return_event(tag)

    def _check_args(self, name, version, rev, stage=None):
        assert_fullname_is_valid(name)
        if stage is not None:
            self.config.assert_stage(stage)
        if version:
            self._check_version(version)
        if not (version is None) ^ (rev is None):
            raise WrongArgs("One and only one of (version, rev) must be specified.")

    @staticmethod
    def _check_version(version):
        if not SemVer.is_valid(version):
            raise WrongArgs(
                f"Version '{version}' is not valid. Example of valid version: 'v1.0.0'"
            )

    def _return_event(self, tag) -> TBaseEvent:  # type: ignore[type-var]
        event = self.check_ref(tag)
        if len(event) > 1:
            raise NotImplementedInGTO("Can't process a tag that caused multiple events")
        return cast(TBaseEvent, event[0])

    @staticmethod
    def _echo_git_suggestion(tag, delete=False):
        echo("To push the changes upstream, run:")
        echo(f"    git push{' --delete ' if delete else ' '}origin {tag}")

    def _delete_tags(self, tags, stdout, push: bool):
        tags = list(tags)
        for tag in tags:
            delete_tag(self.scm, tag)
            if stdout:
                echo(f"Deleted git tag '{tag}'")
            self._push_tag_or_echo_reminder(
                tag_name=tag,
                push=push,
                stdout=stdout,
                delete=True,
            )

    def check_ref(self, ref: str) -> List[BaseEvent]:
        "Find out what was registered/assigned in this ref"
        try:
            name = ""
            tag_name = ref[len("refs/tags/") :] if ref.startswith("refs/tags/") else ref
            if self.scm.get_tag(tag_name):
                # check the ref follows the GTO format
                name = parse_name(tag_name)[NAME]
        except InvalidTagName:
            pass
        if not name:
            logging.info("Provided ref doesn't exist or it is not of GTO format")
            return []
        state = self.get_state()
        return [
            event
            for aname, artifact in state.get_artifacts().items()
            if aname == name
            for event in artifact.get_events()
            # TODO: support matching the shortened commit hashes
            if event.ref == tag_name
        ]

    def find_commit(self, name, version):
        return self.get_state().find_commit(name, version)

    def which(
        self,
        name,
        stage,
        raise_if_not_found=True,
        assignments_per_version=None,
        versions_per_stage=None,
        registered_only=False,
    ):
        """Return stage active in specific stage"""
        return self.get_state().which(
            name,
            stage,
            raise_if_not_found,
            assignments_per_version=assignments_per_version,
            versions_per_stage=versions_per_stage,
            registered_only=registered_only,
        )

    def latest(self, name: str, all: bool = False, registered: bool = True):
        """Return latest active version for artifact"""
        artifact = self.get_state().find_artifact(name)
        if all:
            return artifact.get_versions(include_non_explicit=not registered)
        return artifact.get_latest_version(registered_only=registered)

    def _get_allowed_stages(self):
        return self.config.STAGES

    def _get_used_stages(self):
        return sorted(
            {stage for o in self.get_artifacts().values() for stage in o.unique_stages}
        )

    def get_stages(self, allowed: bool = False, used: bool = False):
        """Return list of stages in the registry.
        If "allowed", return stages that are allowed in config.
        If "used", return stages that were used in registry.
        """
        assert not (allowed and used), """Either "allowed" or "used" can be set"""
        if allowed:
            return self._get_allowed_stages()
        if used:
            return self._get_used_stages()
        # if stages in config are set, return them
        if self._get_allowed_stages() is not None:
            return self._get_allowed_stages()
        # if stages aren't set in config, return those in use
        return self._get_used_stages()

    def _push_tag_or_echo_reminder(
        self, tag_name: str, push: bool, stdout: bool, delete: bool = False
    ) -> None:
        if push or self.cloned:
            if stdout:
                echo(
                    f"Running `git push{' --delete ' if delete else ' '}origin {tag_name}`"
                )
            git_push_tag(
                scm=self.scm,
                tag_name=tag_name,
                delete=delete,
            )
            if stdout:
                echo(
                    f"Successfully {'deleted' if delete else 'pushed'} git tag {tag_name} on remote."
                )
        elif stdout:
            self._echo_git_suggestion(tag_name, delete=delete)



import json
from pathlib import Path

SAMPLE_HTTP_REMOTE_REPO = "https://github.com/iterative/example-gto.git"
SAMPLE_HTTP_REMOTE_REPO_WITHOUT_DOT_GIT_SUFFIX = (
    "https://github.com/iterative/example-gto"
)
SAMPLE_REMOTE_REPO_URL = r"https://github.com/iterative/example-gto-frozen.git"


def get_sample_remote_repo_expected_registry() -> dict:
    with open(
        Path(__file__).parent / "sample_remote_repo_expected_registry.json",
        "r",
        encoding="utf-8",
    ) as f:
        return json.load(f)


def get_sample_remote_repo_expected_history_churn() -> dict:
    with open(
        Path(__file__).parent / "sample_remote_repo_expected_history_churn.json",
        "r",
        encoding="utf-8",
    ) as f:
        return json.load(f)



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Union

from funcy import distinct
from scmrepo.exceptions import RevError
from scmrepo.git import Git

from gto.constants import (
    ARTIFACT,
    ASSIGNMENTS_PER_VERSION,
    COMMIT,
    STAGE,
    VERSION,
    VERSIONS_PER_STAGE,
    VersionSort,
    is_hexsha,
    mark_artifact_unregistered,
    parse_shortcut,
)
from gto.exceptions import NoRepo, NotImplementedInGTO, RefNotFound, WrongArgs
from gto.index import Artifact, RepoIndexManager
from gto.registry import GitRegistry
from gto.tag import parse_name as parse_tag_name


def _is_gto_repo(repo: Union[str, Path, Git]):
    """Check if repo is a gto repo"""
    try:
        with GitRegistry.from_url(repo) as reg:
            return reg.is_gto_repo()
    except NoRepo:
        return False


def _get_state(repo: Union[str, Path, Git]):
    """Show current registry state"""
    with GitRegistry.from_url(repo) as reg:
        return reg.get_state()


def get_stages(repo: Union[str, Path, Git], allowed: bool = False, used: bool = False):
    with GitRegistry.from_url(repo) as reg:
        return reg.get_stages(allowed=allowed, used=used)


def describe(
    repo: Union[str, Path, Git], name: str, rev: Optional[str] = None
) -> Optional[Artifact]:
    """Find enrichments for the artifact"""
    shortcut = parse_shortcut(name)
    if shortcut.shortcut:
        if rev:
            raise WrongArgs("Either specify revision or use naming shortcut.")
        versions = show(repo, name)
        if len(versions) == 0:  # nothing found
            return None
        if len(versions) > 1:
            raise NotImplementedInGTO(
                "Ambiguous naming shortcut: multiple variants found."
            )
        rev = versions[0]["commit_hexsha"]

    with RepoIndexManager.from_url(repo) as index:
        artifact = index.get_commit_index(rev or index.scm.get_rev()).state.get(
            shortcut.name
        )
    return artifact


def register(
    repo: Union[str, Path, Git],
    name: str,
    ref: str,
    version: Optional[str] = None,
    message: Optional[str] = None,
    simple: Optional[bool] = None,
    force: bool = False,
    bump_major: bool = False,
    bump_minor: bool = False,
    bump_patch: bool = False,
    push: bool = False,
    stdout: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    """Register new artifact version"""
    with GitRegistry.from_url(repo) as reg:
        return reg.register(
            name=name,
            rev=ref,
            version=version,
            message=message,
            simple=simple if simple is not None else True,
            force=force,
            bump_major=bump_major,
            bump_minor=bump_minor,
            bump_patch=bump_patch,
            push=push,
            stdout=stdout,
            author=author,
            author_email=author_email,
        )


def assign(
    repo: Union[str, Path, Git],
    name: str,
    stage: str,
    version: Optional[str] = None,
    ref: Optional[str] = None,
    name_version: Optional[str] = None,
    message: Optional[str] = None,
    simple: bool = False,
    force: bool = False,
    push: bool = False,
    skip_registration: bool = False,
    stdout: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    """Assign stage to specific artifact version"""
    with GitRegistry.from_url(repo) as reg:
        return reg.assign(
            name=name,
            stage=stage,
            version=version,
            rev=ref,
            name_version=name_version,
            message=message,
            simple=simple,
            force=force,
            push=push,
            skip_registration=skip_registration,
            stdout=stdout,
            author=author,
            author_email=author_email,
        )


def unassign(
    repo: Union[str, Path, Git],
    name: str,
    ref: Optional[str] = None,
    version: Optional[str] = None,
    stage: Optional[str] = None,
    message: Optional[str] = None,
    stdout: bool = False,
    simple: Optional[bool] = None,
    force: bool = False,
    delete: bool = False,
    push: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    with GitRegistry.from_url(repo) as reg:
        return reg.unassign(
            name=name,
            stage=stage,
            rev=ref,
            version=version,
            message=message,
            stdout=stdout,
            simple=simple if simple is not None else False,
            force=force,
            delete=delete,
            push=push,
            author=author,
            author_email=author_email,
        )


def deregister(
    repo: Union[str, Path, Git],
    name: str,
    ref: Optional[str] = None,
    version: Optional[str] = None,
    message: Optional[str] = None,
    stdout: bool = False,
    simple: Optional[bool] = None,
    force: bool = False,
    delete: bool = False,
    push: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    with GitRegistry.from_url(repo) as reg:
        return reg.deregister(
            name=name,
            rev=ref,
            version=version,
            message=message,
            stdout=stdout,
            simple=simple if simple is not None else True,
            force=force,
            delete=delete,
            push=push,
            author=author,
            author_email=author_email,
        )


def deprecate(
    repo: Union[str, Path, Git],
    name: str,
    ref: Optional[str] = None,
    message: Optional[str] = None,
    stdout: bool = False,
    simple: Optional[bool] = None,
    force: bool = False,
    delete: bool = False,
    push: bool = False,
    author: Optional[str] = None,
    author_email: Optional[str] = None,
):
    with GitRegistry.from_url(repo) as reg:
        return reg.deprecate(
            name=name,
            rev=ref,
            message=message,
            stdout=stdout,
            simple=simple,
            force=force,
            delete=delete,
            push=push,
            author=author,
            author_email=author_email,
        )


def parse_tag(name: str):
    return parse_tag_name(name)


def find_latest_version(
    repo: Union[str, Path, Git],
    name: str,
    all: bool = False,
    registered: bool = True,
):
    """Return latest version for artifact"""
    with GitRegistry.from_url(repo) as reg:
        return reg.latest(name, all=all, registered=registered)


def find_versions_in_stage(
    repo: Union[str, Path, Git],
    name: str,
    stage: str,
    assignments_per_version=ASSIGNMENTS_PER_VERSION,
    versions_per_stage=VERSIONS_PER_STAGE,
    registered_only: bool = False,
):
    """Return version of artifact with specific stage active"""
    with GitRegistry.from_url(repo) as reg:
        return reg.which(
            name,
            stage,
            raise_if_not_found=False,
            assignments_per_version=assignments_per_version,
            versions_per_stage=versions_per_stage,
            registered_only=registered_only,
        )


def check_ref(repo: Union[str, Path, Git], ref: str):
    """Find out what have been registered/assigned in the provided ref"""
    with GitRegistry.from_url(repo) as reg:
        return reg.check_ref(ref)


def show(
    repo: Union[str, Path, Git],
    name: Optional[str] = None,
    truncate_hexsha=False,
    registered_only=False,
    deprecated=False,
    assignments_per_version=ASSIGNMENTS_PER_VERSION,
    versions_per_stage=VERSIONS_PER_STAGE,
    sort=VersionSort.Timestamp,
    table: bool = False,
):
    return (
        _show_versions(
            repo,
            name=name,
            registered_only=registered_only,
            deprecated=deprecated,
            assignments_per_version=assignments_per_version,
            versions_per_stage=versions_per_stage,
            sort=sort,
            table=table,
            truncate_hexsha=truncate_hexsha,
        )
        if name
        else _show_registry(
            repo,
            registered_only=registered_only,
            deprecated=deprecated,
            assignments_per_version=assignments_per_version,
            versions_per_stage=versions_per_stage,
            sort=sort,
            table=table,
            truncate_hexsha=truncate_hexsha,
        )
    )


def _show_registry(
    repo: Union[str, Path, Git],
    registered_only=False,
    deprecated=False,
    assignments_per_version: Optional[int] = None,
    versions_per_stage: Optional[int] = None,
    sort: Optional[VersionSort] = None,
    table: bool = False,
    truncate_hexsha: bool = False,
):
    """Show current registry state"""

    def format_hexsha(hexsha):
        return hexsha[:7] if truncate_hexsha else hexsha

    with GitRegistry.from_url(repo) as reg:
        stages = list(reg.get_stages())
        models_state = {
            o.artifact: {
                "version": format_hexsha(
                    o.get_latest_version(registered_only=True).version
                )
                if o.get_latest_version(registered_only=True)
                else None,
                "stage": {
                    name: ", ".join(
                        [
                            format_hexsha(s.version)
                            for s in o.get_vstages(
                                registered_only=registered_only,
                                assignments_per_version=assignments_per_version,
                                versions_per_stage=versions_per_stage,
                                sort=sort,
                            ).get(name, [])
                        ]
                    )
                    or None
                    for name in stages
                },
                "registered": o.is_registered,
                "active": o.is_active,
            }
            for o in reg.get_artifacts().values()
            if o.is_active or deprecated
        }

    if not table:
        return models_state

    return [
        OrderedDict(
            zip(
                ["name", "latest"] + [f"#{e}" for e in stages],
                [
                    name if d["registered"] else mark_artifact_unregistered(name),
                    d["version"],
                ]
                + [d["stage"][name] for name in stages],
            ),
        )
        for name, d in sorted(models_state.items())
    ], "keys"


def _get_versions(
    artifact,
    deprecated,
    registered_only,
    assignments_per_version,
    versions_per_stage,
    sort,
):
    versions = []
    stages = artifact.get_vstages(
        registered_only=registered_only,
        assignments_per_version=assignments_per_version,
        versions_per_stage=versions_per_stage,
        sort=sort,
    )
    for v in artifact.get_versions(
        active_only=not deprecated,
        include_non_explicit=not registered_only,
        include_discovered=True,
    ):
        v = v.dict_state()
        v["stages"] = [
            vstage.dict_state()
            for vstages in stages.values()
            for vstage in vstages
            if vstage.version == v["version"]
        ]
        if artifact.is_active or deprecated:
            versions.append(v)
    return versions


def _show_versions(  # pylint: disable=too-many-locals
    repo: Union[str, Path, Git],
    name: str,
    ref: Optional[str] = None,
    raw: bool = False,
    registered_only=False,
    deprecated=False,
    assignments_per_version=ASSIGNMENTS_PER_VERSION,
    versions_per_stage=VERSIONS_PER_STAGE,
    sort=VersionSort.Timestamp,
    table: bool = False,
    truncate_hexsha: bool = False,
):
    """List versions of artifact"""

    def format_hexsha(hexsha):
        return hexsha[:7] if truncate_hexsha and is_hexsha(hexsha) else hexsha

    shortcut = parse_shortcut(name)
    if shortcut.shortcut and ref:
        raise WrongArgs("Cannot specify both shortcut and ref")

    with GitRegistry.from_url(repo) as reg:
        if raw:
            return reg.find_artifact(shortcut.name).versions

        artifact = reg.find_artifact(shortcut.name)

        versions = _get_versions(
            artifact,
            deprecated,
            registered_only,
            assignments_per_version,
            versions_per_stage,
            sort,
        )
        if ref:
            try:
                ref_hexsha = reg.scm.resolve_rev(ref)
            except RevError as e:
                raise RefNotFound(ref=ref) from e
            versions = [v for v in versions if v["commit_hexsha"] == ref_hexsha]
        elif shortcut.latest:
            versions = versions[:1]
        elif shortcut.version:
            versions = [v for v in versions if shortcut.version == v["version"]]
        elif shortcut.stage:
            versions = [
                v for v in versions for a in v["stages"] if shortcut.stage == a["stage"]
            ]

    if not table:
        return versions

    first_keys = ["artifact", "version", "stage"]
    versions_ = []
    for v in versions:
        v["artifact"] = (
            v["artifact"]
            if artifact.is_registered
            else mark_artifact_unregistered(v["artifact"])
        )
        v["version"] = format_hexsha(v["version"])
        v["stage"] = ", ".join(
            distinct(  # TODO: remove? no longer necessary
                s["stage"] for s in v["stages"]
            )
        )
        v["commit_hexsha"] = format_hexsha(v["commit_hexsha"])
        v["ref"] = format_hexsha(v["ref"])
        if len(v["registrations"]) > 1:
            raise NotImplementedInGTO(
                "Multiple registrations are not supported currently. How you got in here?"
            )
        for key in list(v.keys()):
            if key not in first_keys + ["created_at", "ref"]:
                del v[key]
        v = OrderedDict(
            [(key, v[key]) for key in first_keys]
            + [(key, v[key]) for key in v if key not in first_keys]
        )
        versions_.append(v)
    return versions_, "keys"


def history(
    repo: Union[str, Path, Git],
    artifact: Optional[str] = None,
    # action: str = None,
    ascending: bool = False,
    table: bool = False,
    truncate_hexsha: bool = False,
):
    with GitRegistry.from_url(repo) as reg:
        artifacts = reg.get_artifacts()

    def format_hexsha(hexsha):
        return hexsha[:7] if truncate_hexsha and is_hexsha(hexsha) else hexsha

    events = [
        OrderedDict(
            timestamp=e.created_at,
            artifact=e.artifact,
            event=type(e).__name__.lower(),
            priority=e.priority,
            version=format_hexsha(e.version) if hasattr(e, "version") else None,
            stage=getattr(e, "stage", None),
            commit=format_hexsha(e.commit_hexsha),
            author=e.author,
            author_email=e.author_email,
            message=e.message,
            ref=format_hexsha(e.ref) if e.ref == e.commit_hexsha else e.ref,
        )
        for o in artifacts.values()
        for e in o.get_events()
        if type(e).__name__.lower() != "commit"
    ]

    events = sorted(
        events,
        key=lambda x: (x["timestamp"], x["priority"]),
    )
    if not ascending:
        events.reverse()
    if artifact:
        events = [event for event in events if event["artifact"] == artifact]
    if not table:
        return events
    keys_order = [
        "timestamp",
        ARTIFACT,
        "event",
        VERSION,
        STAGE,
        COMMIT,
        "ref",
    ]
    keys_order = [c for c in keys_order if any(c in event for event in events)]
    events = [
        OrderedDict((key, event.get(key)) for key in keys_order) for event in events
    ]
    return events, "keys"



__all__ = [
    "BaseModel",
    "BaseSettings",
    "ValidationError",
    "parse_obj_as",
    "validator",
    "InitSettingsSource",
]


try:
    from pydantic.v1 import (
        BaseModel,
        BaseSettings,
        ValidationError,
        parse_obj_as,
        validator,
    )
    from pydantic.v1.env_settings import InitSettingsSource
except ImportError:
    from pydantic import (  # type: ignore[no-redef,assignment]
        BaseModel,
        BaseSettings,
        ValidationError,
        parse_obj_as,
        validator,
    )
    from pydantic.env_settings import (  # type: ignore[no-redef,assignment]
        InitSettingsSource,
    )



from unittest.mock import MagicMock

import pytest
from pytest_test_utils import TmpDir
from scmrepo.exceptions import SCMError
from scmrepo.git import Git, SyncStatus

from gto.exceptions import GTOException
from gto.git_utils import git_add_and_commit_all_changes, git_push_tag


@pytest.fixture(name="mocked_scm")
def _mocked_scm() -> MagicMock:
    return MagicMock(
        root_dir="git_repo_path",
        validate_git_remote=MagicMock(return_value=True),
    )


def test_git_push_tag(mocked_scm):
    tag_name = "test_tag"
    mocked_scm.push_refspecs = MagicMock(
        return_value={f"refs/tag/{tag_name}": SyncStatus.SUCCESS}
    )
    git_push_tag(mocked_scm, tag_name=tag_name)
    mocked_scm.push_refspecs.assert_called_once_with(
        "origin", "refs/tags/test_tag:refs/tags/test_tag"
    )


def test_git_push_tag_with_delete(mocked_scm):
    tag_name = "test_tag"
    mocked_scm.push_refspecs = MagicMock(
        return_value={f"refs/tag/{tag_name}": SyncStatus.SUCCESS}
    )
    git_push_tag(mocked_scm, tag_name=tag_name, delete=True)
    mocked_scm.push_refspecs.assert_called_once_with("origin", ":refs/tags/test_tag")


def test_git_push_tag_error(mocked_scm):
    mocked_scm.push_refspecs = MagicMock(side_effect=SCMError())
    tag_name = "test_tag"
    with pytest.raises(GTOException) as error:
        git_push_tag(mocked_scm, tag_name=tag_name)

    assert f"git push origin {tag_name}" in error.value.msg
    assert (
        "Make sure your local repository is in sync with the remote" in error.value.msg
    )


@pytest.mark.usefixtures("repo_with_commit")
def test_git_add_and_commit_all_changes_if_files_not_changed_then_no_new_commit(
    scm: Git,
):
    rev = scm.get_rev()
    git_add_and_commit_all_changes(
        scm=scm,
        message="test message",
    )
    assert scm.get_rev() == rev


@pytest.mark.usefixtures("repo_with_commit")
def test_git_add_and_commit_all_changes_if_tracked_file_is_changed_then_new_commit(
    tmp_dir: TmpDir,
    scm: Git,
):
    rev = scm.get_rev()
    tmp_dir.gen("some-tracked-file", "test data")
    git_add_and_commit_all_changes(
        scm,
        message="test message",
    )
    assert scm.get_rev() != rev
    fs = scm.get_fs("HEAD")
    with fs.open("some-tracked-file", "r", encoding="utf-8") as f:
        assert f.read() == "test data"


@pytest.mark.usefixtures("repo_with_commit")
def test_git_commit_specific_files_if_untracked_file_is_changed_then_new_commit(
    tmp_dir: TmpDir,
    scm: Git,
):
    rev = scm.get_rev()
    tmp_dir.gen("untracked-file", "test data")
    git_add_and_commit_all_changes(
        scm,
        message="test message",
    )
    assert scm.get_rev() != rev
    fs = scm.get_fs("HEAD")
    with fs.open("untracked-file", "r", encoding="utf-8") as f:
        assert f.read() == "test data"



import pytest

from gto.constants import (
    check_string_is_valid,
    fullname_in_tag_re,
    fullname_re,
)


@pytest.mark.parametrize(
    "name",
    [
        "1",
        "m",
        "nn",
        "m1",
        "1nn",
        "model-prod",
        "model-prod-v1",
        "model_prod_v1",
        "dvclive/model",
    ],
)
def test_check_name_is_valid(name):
    assert check_string_is_valid(name)


@pytest.mark.parametrize(
    "name",
    [
        "",
        "m/",
        "/m",
        "###",
        "@@@",
        "a model",
        "_model",
        "-model",
        "model-",
        "model@1",
        "model#1",
        "@namespace/model",
        "DVCLive/Model",
        "model A",
    ],
)
def test_check_name_is_invalid(name):
    assert not check_string_is_valid(name)


@pytest.mark.parametrize(
    "name",
    [
        "model",
        "dvclive:model",
        "some/folder:some/model",
    ],
)
def test_check_fullname_is_valid(name):
    assert check_string_is_valid(name, regex=fullname_re)


@pytest.mark.parametrize(
    "name",
    [
        "model",
        "dvclive=model",
        "some/folder=some/model",
    ],
)
def test_check_fullname_in_tag_is_valid(name):
    assert check_string_is_valid(name, regex=fullname_in_tag_re)



import re
from enum import Enum
from typing import Optional

from gto.exceptions import ValidationError

from ._pydantic import BaseModel

COMMIT = "commit"
REF = "ref"
TAG = "tag"

ARTIFACT = "artifact"
ACTION = "action"
TYPE = "type"
NAME = "name"
PATH = "path"
VERSION = "version"
STAGE = "stage"
COUNTER = "counter"


class Action(Enum):
    CREATE = "create"
    DEPRECATE = "deprecate"
    DEREGISTER = "deregister"
    REGISTER = "register"
    ASSIGN = "assign"
    UNASSIGN = "unassign"


SEPARATOR_IN_NAME = ":"
SEPARATOR_IN_TAG = "="


def name_to_tag(value):
    return value.replace(SEPARATOR_IN_NAME, SEPARATOR_IN_TAG)


def tag_to_name(value):
    return value.replace(SEPARATOR_IN_TAG, SEPARATOR_IN_NAME)


dirname = "[a-z0-9-_./]+"  # improve?
name = r"[a-z0-9]([a-z0-9-/_]*[a-z0-9])?"
semver = r"(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?"
counter = "?P<counter>[0-9]+"
name_re = re.compile(f"^{name}$")
fullname = f"(?P<dirname>{dirname}{SEPARATOR_IN_NAME})?(?P<name>{name})"  # add test to check ":" is here only once?
fullname_in_tag = name_to_tag(fullname)
fullname_re = re.compile(f"^{fullname}$")
fullname_in_tag_re = re.compile(f"^{fullname_in_tag}$")

tag_re = re.compile(
    f"^(?P<artifact>{fullname_in_tag})(((#(?P<stage>{name})|@(?P<version>v{semver}))(?P<cancel>!?))|@((?P<deprecated>deprecated)|(?P<created>created)))(#({counter}))?$"
)
shortcut_re = re.compile(
    f"^(?P<artifact>{fullname})(#(?P<stage>{name})|@(?P<version>latest|greatest|v{semver}))$"
)
git_hexsha_re = re.compile(r"^[0-9a-fA-F]{40}$")


def is_hexsha(value):
    return bool(git_hexsha_re.search(value))


def check_string_is_valid(value, regex=name_re):
    return bool(regex.search(value))


def assert_name_is_valid(value):
    if not check_string_is_valid(value, regex=name_re):
        raise ValidationError(
            f"Invalid value '{value}'. Only lowercase english letters, , '-', '/' are allowed."
            "Value must be of len >= 2, must with a letter and end with a letter or a number."
        )


def assert_fullname_is_valid(value):
    if not check_string_is_valid(value, regex=fullname_re):
        # fix error message to be regex-specific
        raise ValidationError(
            f"Invalid value '{value}'. Only lowercase english letters, , '-', '/' are allowed."
            "Value must be of len >= 2, must start with a letter and end with a letter or a number."
        )


class Shortcut(BaseModel):
    name: str
    stage: Optional[str] = None
    version: Optional[str] = None
    latest: bool = False
    shortcut: bool = False


def parse_shortcut(value):
    match = re.search(shortcut_re, value)
    if match:
        value = match["artifact"]
        if match["stage"]:
            assert_name_is_valid(match["stage"])
    latest = bool(match and (match["version"] in ("latest", "greatest")))
    return Shortcut(
        name=value,
        stage=match["stage"] if match and match["stage"] else None,
        version=match["version"] if match and match["version"] and not latest else None,
        latest=latest,
        shortcut=bool(match),
    )


def mark_artifact_unregistered(artifact_name):
    return f"*{artifact_name}"


class VersionSort(Enum):
    SemVer = "semver"
    Timestamp = "timestamp"


ASSIGNMENTS_PER_VERSION = -1
VERSIONS_PER_STAGE = 1



import os
import re
from datetime import datetime
from enum import Enum
from typing import FrozenSet, Iterable, Optional, Union

from scmrepo.exceptions import RevError
from scmrepo.git import Git, GitTag

from ._pydantic import BaseModel
from .base import (
    Artifact,
    Assignment,
    BaseManager,
    BaseRegistryState,
    Deprecation,
    Deregistration,
    Registration,
    Unassignment,
)
from .constants import (
    ACTION,
    COUNTER,
    NAME,
    STAGE,
    TAG,
    VERSION,
    Action,
    name_to_tag,
    tag_re,
    tag_to_name,
)
from .exceptions import (
    InvalidTagName,
    MissingArg,
    NotImplementedInGTO,
    RefNotFound,
    TagExists,
    TagNotFound,
    UnknownAction,
)

COUNT_DELIMITER = "#"

TagTemplates = {
    # Action.CREATE: "{artifact}@",
    Action.DEPRECATE: "{artifact}@deprecated",
    Action.REGISTER: "{artifact}@{version}",
    Action.DEREGISTER: "{artifact}@{version}!",
    Action.ASSIGN: "{artifact}#{stage}",
    Action.UNASSIGN: "{artifact}#{stage}!",
}


def name_tag(
    action: Action,
    artifact: str,
    version: Optional[str] = None,
    stage: Optional[str] = None,
    scm: Optional[Git] = None,
    simple: bool = False,
):
    if action not in TagTemplates:
        raise UnknownAction(action=action)

    tag = TagTemplates[action].format(
        artifact=name_to_tag(artifact), version=version, stage=stage
    )
    if simple:
        return tag
    if scm is None:
        raise MissingArg(arg="scm")
    counter = 0
    for t in scm.list_tags():
        parsed = parse_name(t, raise_on_fail=False)  # type: ignore
        if (
            parsed
            and parsed[NAME] == artifact
            and COUNTER in parsed
            and parsed[COUNTER] > counter
        ):
            counter = parsed[COUNTER]
    return f"{tag}{COUNT_DELIMITER}{counter+1}"


def parse_name(name: str, raise_on_fail: bool = True):
    match = re.search(tag_re, name)
    if raise_on_fail and not match:
        raise InvalidTagName(name)
    if match:
        parsed = {NAME: tag_to_name(match["artifact"])}
        if match["deprecated"]:
            parsed[ACTION] = Action.DEPRECATE
        if match[VERSION]:
            parsed[VERSION] = match[VERSION]
            parsed[ACTION] = (
                Action.DEREGISTER if match["cancel"] == "!" else Action.REGISTER
            )
        if match[STAGE]:
            parsed[STAGE] = match[STAGE]
            parsed[ACTION] = (
                Action.UNASSIGN if match["cancel"] == "!" else Action.ASSIGN
            )
        if match[COUNTER]:
            parsed[COUNTER] = int(match[COUNTER])
        return parsed
    return {}


class NAME_REFERENCE(Enum):
    TAG = TAG
    NAME = NAME


def parse_name_reference(name: str):
    parsed = parse_name(name, raise_on_fail=False)
    if not parsed:
        return NAME_REFERENCE.NAME, name
    return NAME_REFERENCE.TAG, parsed


class Tag(BaseModel):
    action: Action
    name: str
    version: Optional[str]
    stage: Optional[str]
    created_at: datetime
    tag: GitTag

    class Config:
        arbitrary_types_allowed = True


def parse_tag(tag: GitTag):
    return Tag(
        tag=tag,
        created_at=datetime.fromtimestamp(tag.tag_time),
        **parse_name(tag.name),
    )


def find(
    action: Optional[Union[Action, FrozenSet[Action]]] = None,
    name: Optional[str] = None,
    version: Optional[str] = None,
    stage: Optional[str] = None,
    scm: Optional[Git] = None,
    sort: str = "by_time",
    tags: Optional[Iterable[GitTag]] = None,
):
    if isinstance(action, Action):
        action = frozenset([action])
    if scm is None:
        raise MissingArg(arg="scm")
    result = []
    tag_names = [t.name for t in tags] if tags else scm.list_tags()
    for t in tag_names:
        try:
            parsed = parse_name(t)
        except InvalidTagName:
            continue
        if (  # pylint: disable=too-many-boolean-expressions
            parsed
            and (not action or parsed[ACTION] in action)
            and (not name or parsed.get(NAME) == name)
            and (not version or parsed.get(VERSION) == version)
            and (not stage or parsed.get(STAGE) == stage)
        ):
            tag = scm.get_tag(t)
            # remove lightweight tags
            if isinstance(tag, GitTag):
                result.append(tag)
    if sort == "by_time":
        return sorted(result, key=lambda t: t.tag_time)
    raise NotImplementedError(f"Unknown sort: {sort}")


def create_tag(  # pylint: disable=too-many-branches
    scm: Git,
    name: str,
    rev: str,
    message: str,
    tagger: Optional[str] = None,
    tagger_email: Optional[str] = None,
):
    try:
        rev = scm.resolve_rev(rev)
    except RevError as e:
        raise RefNotFound(ref=rev) from e
    if scm.get_tag(name):
        raise TagExists(name=name)

    if tagger:
        orig_name: Optional[str] = os.environ.get("GIT_COMMITTER_NAME")
        os.environ["GIT_COMMITTER_NAME"] = tagger
    else:
        orig_name = None
    if tagger_email:
        orig_email: Optional[str] = os.environ.get("GIT_COMMITTER_EMAIL")
        os.environ["GIT_COMMITTER_EMAIL"] = tagger_email
    else:
        orig_email = None
    try:
        scm.tag(name, target=rev, annotated=True, message=message)
    finally:
        if tagger:
            if orig_name:
                os.environ["GIT_COMMITTER_NAME"] = orig_name
            else:
                del os.environ["GIT_COMMITTER_NAME"]
        if tagger_email:
            if orig_email:
                os.environ["GIT_COMMITTER_EMAIL"] = orig_email
            else:
                del os.environ["GIT_COMMITTER_EMAIL"]


def delete_tag(scm: Git, name: str):
    ref = f"refs/tags/{name}"
    if not scm.get_ref(ref):
        raise TagNotFound(name=name)
    scm.remove_ref(ref)


def index_tag(artifact: Artifact, tag: GitTag) -> Artifact:
    event: Union[Deprecation, Registration, Deregistration, Assignment, Unassignment]
    mtag = parse_tag(tag)
    if mtag.action == Action.REGISTER:
        event = Registration(
            artifact=mtag.name,
            version=mtag.version,
            created_at=mtag.created_at,
            author=tag.tagger_name,
            author_email=tag.tagger_email,
            message=tag.message.strip(),
            commit_hexsha=tag.target,
            tag=tag.name,
        )
    elif mtag.action == Action.DEREGISTER:
        event = Deregistration(
            artifact=mtag.name,
            version=mtag.version,
            created_at=mtag.created_at,
            author=tag.tagger_name,
            author_email=tag.tagger_email,
            message=tag.message.strip(),
            commit_hexsha=tag.target,
            tag=tag.name,
        )
    elif mtag.action in (Action.ASSIGN, Action.UNASSIGN):
        version = artifact.find_version(
            commit_hexsha=tag.target, create_new=True
        ).version  # type: ignore
        if mtag.action == Action.ASSIGN:
            event = Assignment(
                artifact=mtag.name,
                version=version,
                stage=mtag.stage,
                created_at=mtag.created_at,
                author=tag.tagger_name,
                author_email=tag.tagger_email,
                message=tag.message.strip(),
                commit_hexsha=tag.target,
                tag=tag.name,
            )
        else:
            event = Unassignment(
                artifact=mtag.name,
                version=version,
                stage=mtag.stage,
                created_at=mtag.created_at,
                author=tag.tagger_name,
                author_email=tag.tagger_email,
                message=tag.message.strip(),
                commit_hexsha=tag.target,
                tag=tag.name,
            )
    elif mtag.action == Action.DEPRECATE:
        event = Deprecation(
            artifact=mtag.name,
            created_at=mtag.created_at,
            author=tag.tagger_name,
            author_email=tag.tagger_email,
            message=tag.message.strip(),
            commit_hexsha=tag.target,
            tag=tag.name,
        )
    artifact.add_event(event)
    return artifact


class TagManager(BaseManager):  # pylint: disable=abstract-method
    def update_state(self, state: BaseRegistryState) -> BaseRegistryState:
        # tags are sorted and then indexed by timestamp
        # this is important to check that history is not broken
        for tag in find(scm=self.scm, action=self.actions):
            state.update_artifact(
                index_tag(
                    state.find_artifact(parse_name(tag.name)["name"], create_new=True),
                    tag,
                )
            )
        return state


class TagArtifactManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.CREATE, Action.DEPRECATE))

    def create(self):  # pylint: disable=no-self-use
        raise NotImplementedInGTO(
            "If you want to create artifact, register a version or assign a stage for it"
        )

    def deprecate(
        self,
        name,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ):
        tag = name_tag(
            Action.DEPRECATE,
            name,
            scm=self.scm,
            simple=simple,
        )
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag


class TagVersionManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.REGISTER, Action.DEREGISTER))

    def register(
        self,
        name,
        version,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ):
        tag = name_tag(
            Action.REGISTER, name, version=version, scm=self.scm, simple=simple
        )
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag

    def deregister(
        self,
        name,
        version,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ):
        tag = name_tag(
            Action.DEREGISTER, name, version=version, scm=self.scm, simple=simple
        )
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag


class TagStageManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.ASSIGN, Action.UNASSIGN))

    def assign(
        self,
        name,
        stage,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> str:
        tag = name_tag(Action.ASSIGN, name, stage=stage, scm=self.scm, simple=simple)
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag

    def unassign(
        self,
        name,
        stage,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> str:
        tag = name_tag(Action.UNASSIGN, name, stage=stage, scm=self.scm, simple=simple)
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag



from abc import ABC, abstractmethod
from functools import lru_cache
from importlib import import_module
from typing import Dict, Optional, Type, Union

import entrypoints
from scmrepo.git import Git

from ._pydantic import BaseModel

ENRICHMENT_ENTRYPOINT = "gto.enrichment"


def import_string(path):
    split = path.split(".")
    module_name, object_name = ".".join(split[:-1]), split[-1]
    mod = import_module(module_name)
    try:
        return getattr(mod, object_name)
    except AttributeError as e:
        raise ImportError(f"No object {object_name} in module {module_name}") from e


class EnrichmentInfo(BaseModel, ABC):
    source: str

    @abstractmethod
    def get_object(self) -> BaseModel:
        raise NotImplementedError

    def get_dict(self):
        return self.get_object().dict()

    @abstractmethod
    def get_human_readable(self) -> str:
        raise NotImplementedError


class EnrichmentReader(BaseModel, ABC):
    source: str

    @abstractmethod
    def describe(
        self, url_or_scm: Union[str, Git], obj: str, rev: Optional[str]
    ) -> Optional[EnrichmentInfo]:
        raise NotImplementedError


@lru_cache()
def _find_enrichments():
    eps = entrypoints.get_group_named(ENRICHMENT_ENTRYPOINT)
    return {k: ep.load() for k, ep in eps.items()}


@lru_cache()
def find_enrichments() -> Dict[str, EnrichmentReader]:
    enrichments = _find_enrichments()
    res = {}
    for name, e in enrichments.items():
        # if isinstance(e, type) and issubclass(e, Enrichment) and not e.__fields_set__:
        if isinstance(e, type) and issubclass(e, EnrichmentReader):
            res[name] = e()  # type: ignore[call-arg]
        if isinstance(e, EnrichmentReader):
            res[name] = e
    return res


@lru_cache()
def find_enrichment_types() -> Dict[str, Type[EnrichmentReader]]:
    enrichments = _find_enrichments()
    return {
        k: e
        for k, e in enrichments.items()
        if isinstance(e, type) and issubclass(e, EnrichmentReader)
    }



from pathlib import Path

from setuptools import find_packages, setup

install_requires = [
    "scmrepo>=1.3.1,<2",
    "typer>=0.4.1",
    "rich",
    # pydantic.v1.parse_obj is broken in ==2.0.0:
    # https://github.com/pydantic/pydantic/issues/6361
    "pydantic>=1.9.0,<3,!=2.0.0",
    "ruamel.yaml",
    "semver>=3.0.0",
    "entrypoints",
    "tabulate>=0.8.10",
    "funcy",
]


tests = [
    "pytest",
    "pytest-cov",
    "pytest-lazy-fixture==0.6.3",
    "pytest-mock",
    "pytest-test-utils",
    "pylint==2.17.5",
    # we use this to suppress some messages in tests, eg: foo/bar naming,
    # and, protected method calls in our tests
    "pylint-plugin-utils",
    "freezegun",
    "types-freezegun",
]


setup_args = {
    "name": "gto",
    "use_scm_version": True,
    "setup_requires": ["setuptools_scm", "fastentrypoints>=0.12"],
    "description": "Version and deploy your models following GitOps principles",
    "long_description": (Path(__file__).parent / "README.md").read_text(
        encoding="utf8"
    ),
    "long_description_content_type": "text/markdown",
    "author": "Alexander Guschin",
    "author_email": "aguschin@iterative.ai",
    "download_url": "https://github.com/iterative/gto",
    "license": "Apache License 2.0",
    "install_requires": install_requires,
    "extras_require": {"tests": tests},
    "keywords": "git repo repository artifact registry developer-tools collaboration",
    "python_requires": ">=3.8",
    "classifiers": [
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    "packages": find_packages(exclude=["tests"]),
    "include_package_data": True,
    "entry_points": {
        "console_scripts": ["gto = gto.cli:app"],
        "gto.enrichment": [
            "gto = gto.index:GTOEnrichment",
        ],
    },
    "zip_safe": False,
}

if __name__ == "__main__":
    setup(**setup_args)



import logging
import os.path
from contextlib import contextmanager
from itertools import chain
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Union

from scmrepo.exceptions import SCMError
from scmrepo.git import Git, SyncStatus

from gto.config import RegistryConfig
from gto.exceptions import GTOException, NoRepo, WrongArgs
from gto.ui import echo


class RemoteRepoMixin:
    @classmethod
    @contextmanager
    def from_scm(
        cls,
        scm: Git,
        cloned: bool = False,
        config: Optional[RegistryConfig] = None,
    ):
        """
        `cloned` - scm is a remote repo that was cloned locally into a tmp
                   directory to be used for the duration of the context manager.
                   Means that we push tags and changes back to the remote repo.
        """
        raise NotImplementedError()

    @classmethod
    @contextmanager
    def from_url(
        cls,
        url_or_scm: Union[str, Path, Git],
        config: Optional[RegistryConfig] = None,
        branch=None,
    ):
        if isinstance(url_or_scm, Git):
            with cls.from_scm(url_or_scm) as obj:
                yield obj
            return
        url_or_scm = str(url_or_scm)
        if os.path.exists(url_or_scm):
            scm = Git(url_or_scm)
            try:
                scm.dir
            except SCMError as e:
                scm.close()
                raise NoRepo(url_or_scm) from e
            if branch:
                raise WrongArgs("branch can only be set for remote repos")
            try:
                with cls.from_scm(scm=scm, config=config) as obj:
                    yield obj
            finally:
                scm.close()
        else:
            with cloned_git_repo(url_or_scm) as scm:
                if branch:
                    scm.checkout(branch)
                with cls.from_scm(scm=scm, config=config, cloned=True) as obj:
                    yield obj

    def _call_commit_push(
        self,
        func,
        commit=False,
        commit_message=None,
        push=False,
        stdout=False,
        **kwargs,
    ):
        if not (commit or push):
            return func(**kwargs, stdout=stdout)
        stashed = set()
        stashed.update(self.scm.status())
        with self.scm.stash_workspace(include_untracked=True):
            result = func(**kwargs, stdout=stdout)
            if any(stashed.intersection(files) for files in self.scm.status()):
                _reset_repo_to_head(self.scm)
                raise GTOException(
                    msg="The command would have changed files that were not committed, "
                    "automated committing is not possible.\n"
                    "Suggested action: Commit the changes and re-run this command."
                )
            git_add_and_commit_all_changes(
                self.scm,
                message=commit_message,
            )
            if push:
                self.scm.push()
            if stdout:
                echo(
                    "Running `git commit`"
                    if not push
                    else "Running `git commit` and `git push`"
                    "\nSuccessfully pushed a new commit to remote."
                )
        return result


@contextmanager
def cloned_git_repo(url: str) -> Git:
    with TemporaryDirectory() as tmp_dir:
        logging.debug("create temporary directory %s", tmp_dir)
        scm = Git.clone(url, tmp_dir)
        try:
            scm.dir
        except SCMError as e:
            scm.close()
            raise NoRepo(url) from e
        try:
            yield scm
        finally:
            scm.close()


def git_push_tag(
    scm: Git,
    tag_name: str,
    delete: bool = False,
    remote_name: str = "origin",
) -> None:
    ref = f"refs/tags/{tag_name}"
    src = "" if delete else ref
    refspec = f"{src}:{ref}"
    logging.debug(
        "push %s tag %s from directory %s to remote %s",
        "--delete" if delete else "",
        tag_name,
        scm.root_dir,
        remote_name,
    )
    try:
        result = scm.push_refspecs(remote_name, refspec)
    except SCMError as e:
        raise GTOException(
            msg=f"The command `git push {remote_name} {tag_name}` failed. "
            f"Make sure your local repository is in sync with the remote."
        ) from e
    for _ref, status in result.items():
        if status == SyncStatus.DIVERGED:
            raise GTOException(
                msg=f"The command `git push {remote_name} {tag_name}` failed. "
                f"Make sure your local repository is in sync with the remote."
            )


def git_add_and_commit_all_changes(scm: Git, message: str) -> None:
    staged, unstaged, untracked = scm.status()
    if staged or unstaged or untracked:
        files = list(chain(staged, unstaged, untracked))
        logging.debug("Adding to the index the untracked files %s", untracked)
        logging.debug("Add and commit changes to files %s", files)
        scm.add(files)
        scm.commit(message)


def _reset_repo_to_head(scm: Git) -> None:
    if scm.stash.push(include_untracked=True):
        scm.stash.drop()
